from __future__ import annotations

from datetime import datetime, date
from typing import Iterator

from weasel_client.resources.Resource import Resource


class DateStatistics(Resource):
    """
    Resource class for DateStatistics
    """
    _date: str

    def __init__(
            self,
            APIClient,
            date: str
    ):
        super().__init__(APIClient=APIClient)
        self._date = date

    def date(self) -> datetime:
        """
        Returns the date-attribute
        """
        return datetime.fromisoformat(self._date)

    def results(self) -> Iterator[Resource]:
        """
        Returns /installations/<date>/ for the object's date
        """
        yield from self._client.installations_list(date=self._date)

    @staticmethod
    def from_date(APIClient, date: date) -> DateStatistics:
        """
        Creates a DateStatistics-object with a given date
        :param APIClient: client to for API-Requests
        :param date: date for the DateStatistics-object
        """
        return DateStatistics(APIClient=APIClient, date=date.isoformat())



class ConnectionException(Exception):
    """
    Exception for failed connections to the WEASEL-API
    """
    pass


class InvalidTokenException(Exception):
    """
    Exception for authorization errors based on a false token
    """
    pass



from datetime import datetime
from typing import Optional

from weasel_client.resources.Resource import Resource


class ReleaseCounter(Resource):
    """
    Resource class for ReleaseCounter
    """
    _release: int
    _observed_day: str
    _count: int

    def __init__(
            self,
            APIClient,
            release: int,
            observed_day: str,
            count: int
    ):
        super().__init__(APIClient=APIClient)
        self._release = release
        self._observed_day = observed_day
        self._count = count

    def release(self) -> Optional[Resource]:
        """
        Returns the associated release
        """
        return self._client.release_detail(pk=self._release)

    def observed_day(self) -> datetime:
        """
        Returns the associated date
        """
        return datetime.fromisoformat(self._observed_day)

    def count(self) -> int:
        """
        Returns the count
        :return:
        """
        return self._count



from pprint import pprint
from typing import Optional, List, Dict, Iterator

from weasel_client.ConnectionManager import ConnectionManager
from weasel_client.Exceptions import ConnectionException
from weasel_client.resources import Technology, Release, SourceFile, Vulnerability, DateStatistics, \
    ReleaseCounter, \
    ReleaseSourcefile


class APIClient:
    """
    Interfaces directly with every API-Endpoint of the WEASEL-API
    """
    __connection: ConnectionManager = None
    __technologies: Dict[str, Technology] = None

    def __init__(self, token: str, url: str = "https://weasel.cs.uni-bonn.de/api/"):
        self.__connection = ConnectionManager(token=token, url=url)

    def connected(self):
        """
	    Checks if the APIClient is connected
	    """
        return self.__connection.connected()

    def url(self):
        """
        Returns the API base url used for the connection
        """
        return self.__connection.url()

    '''
        Client Endpoints of API Root Level
    '''

    def releases(self) -> Iterator[Release]:
        """
        Alias for release_list
        """
        yield from self.release_list()

    def technologies(self) -> Dict[str, Technology]:
        """
        Alias for technology_list
        """
        return self.technology_list()

    def vulnerabilities(self) -> Iterator[Vulnerability]:
        """
        Alias for vulnerability_list
        """
        yield from self.vulnerability_list()

    # Alias for installations_day_list
    def installations(self) -> Iterator[DateStatistics]:
        """
        Alias for installations_day_list
        """
        yield from self.installations_day_list()

    '''
        Convenience Functions
    '''

    def technology(self, name: str) -> Technology:
        """
        Basically a technology-detail Endpoint (non-existent in API)
        :param name: Name of the technology to fetch
        """
        for tech_name in self.technology_list().keys():
            if tech_name.lower() == name.lower():
                return self.__technologies[tech_name]

    def release(self, pk: int) -> Optional[Release]:
        """
        Alias for release_detail
        :param pk: Private key of release to get
        """
        return self.release_detail(pk=pk)

    '''
        Client Endpoints per API view
    '''

    def release_list(self) -> Iterator[Release]:
        """
        Fetcher for releases/
        """
        yield from self.__release_serializer_paginated_fetcher("releases/")

    def release_detail(self, pk: int) -> Optional[Release]:
        """
        Fetcher for releases/<int:pk>/
        :param pk: Private key of release to get
        """
        try:
            release_desc = self.__connection.get(request="releases/" + str(pk) + "/").json()
        except ConnectionException:
            return None
        if release_desc['prev_release'] is not None:
            prev_release = self.__fetch_end_of_path_id(release_desc['prev_release'])
        else:
            prev_release = None
        return Release(
            APIClient=self,
            id=release_desc['id'],
            url=release_desc['url'],
            technology=self.technology(release_desc['technology']),
            version=release_desc['version'],
            prev_release=prev_release,
            published=release_desc['published']
        )

    def sourcecode_zip(self, pk: int):
        """
        Fetcher for releases/<int:pk>/zip/
        :param pk: Private key of associated release
        """
        try:
            return self.__connection.get(request="releases/" + str(pk) + "/zip/")
        except ConnectionException:
            return None

    def sourcefile_list(self, pk: int) -> Iterator[SourceFile]:
        """
        Fetcher for releases/<int:pk>/filelist/
        :param pk: Private key of associated release
        """
        yield from self.__sourcefile_serializer_paginated_fetcher(
            url="releases/" + str(pk) + "/filelist/")

    def release_vulnerabilities(self, pk: int) -> Iterator[Vulnerability]:
        """
        Fetcher for releases/<int:pk>/vulnerabilities/
        :param pk: Private key of associated release
        """
        yield from self.__vulnerability_serializer_paginated_fetcher(
            url="releases/" + str(pk) + "/vulnerabilities/")

    def release_installations(self, pk: int) -> Iterator[ReleaseCounter]:
        """
        Fetcher for releases/<int:pk>/installations/
        :param pk: Private key of associated release
        """
        yield from self.__release_counter_release_serializer_paginated_fetcher(
            url="releases/" + str(pk) + "/installations/", release=pk)

    def release_tech_list(self, tech_name: str) -> Iterator[Release]:
        """
        Fetcher for releases/<str:tech_name>/
        :param tech_name: Name of the associated technology
        """
        yield from self.__release_serializer_paginated_fetcher(url="releases/" + tech_name + "/")

    def technology_list(self) -> Dict[str, Technology]:
        """
        Fetcher for technologies/
        """
        if self.__technologies is None:
            self.__technologies = self.__fetch_technologies()
        return self.__technologies

    def sourcefile_byhash(self, type: str, hash: str) -> Iterator[SourceFile]:
        """
        Fetcher for file/<str:type>/<str:hash>/
        :param type: md5|sha1|sha256|ssdeep
        :param hash: hash in HEX-Format
        """
        yield from self.__release_sourcefile_serializer_paginated_fetcher(
            url="file/" + type + "/" + hash + "/")

    def sourcefile_detail(self, pk: int) -> Optional[SourceFile]:
        """
        Fetcher for file/<int:pk>/
        :param pk: Private key of file to get
        """
        try:
            sourcefile_desc = self.__connection.get(request="file/" + str(pk) + "/").json()
        except ConnectionException:
            return None
        return SourceFile(
            APIClient=self,
            id=sourcefile_desc['id'],
            url=sourcefile_desc['url'],
            hash_sha1=sourcefile_desc['hash_sha1'],
            hash_sha256=sourcefile_desc['hash_sha256'],
            hash_md5=sourcefile_desc['hash_md5'],
            hash_ssdeep=sourcefile_desc['hash_ssdeep']
        )

    def sourcefile_download(self, pk: int):
        """
        Fetcher for file/<int:pk>/download/
        :param pk: Private key of associated file
        """
        try:
            return self.__connection.get(request="file/" + str(pk) + "/download/")
        except ConnectionException:
            return None

    def sourcefile_releases(self, pk: int) -> Iterator[Release]:
        """
        Fetcher for file/<int:pk>/releases/
        :param pk: Private key of associated file
        """
        yield from self.__release_sourcefile_serializer_paginated_fetcher(
            url="file/" + str(pk) + "/releases/")

    def vulnerability_list(self) -> Iterator[Vulnerability]:
        """
        Fetcher for vulnerability/
        """
        yield from self.__vulnerability_serializer_paginated_fetcher(url="vulnerability/")

    def vulnerability_detail(self, pk: int) -> Optional[Vulnerability]:
        """
        Fetcher for vulnerability/<int:pk>/
        :param pk: Private key of vulnerability to get
        """
        try:
            vuln_desc = self.__connection.get(request="vulnerability/" + str(pk) + "/").json()
        except ConnectionException:
            return None
        return Vulnerability(
            APIClient=self,
            id=vuln_desc['id'],
            url=vuln_desc['url'],
            cve_number=vuln_desc['CVE_number'],
            publish_date=vuln_desc['publish_date'],
            last_modified=vuln_desc['last_modified'],
            title=vuln_desc['title'],
            type=vuln_desc['type'],
            references=vuln_desc['references'],
            fishy=vuln_desc['fishy']
        )

    def vulnerability_found_in(self, pk: int) -> Iterator[Release]:
        """
        Fetcher for vulnerability/<int:pk>/found_in/
        :param pk: Private key of associated vulnerability
        """
        yield from self.__release_serializer_paginated_fetcher(
            url="vulnerability/" + str(pk) + "/found_in/")

    def vulnerability_fixed_by(self, pk: int) -> Iterator[Release]:
        """
        Fetcher for vulnerability/<int:pk>/fixed_by/
        :param pk: Private key of associated vulnerability
        """
        yield from self.__release_serializer_paginated_fetcher(
            url="vulnerability/" + str(pk) + "/fixed_by/")

    def installations_day_list(self) -> Iterator[DateStatistics]:
        """
        Fetcher for installations/
        """
        yield from self.__installations_day_serializer_paginated_fetcher("installations/")

    def installations_list(self, date: str) -> Iterator[ReleaseCounter]:
        """
        Fetcher for installations/<str:date>/
        :param date: Associated date in ISO format
        :return:
        """
        yield from self.__release_counter_date_serializer_paginated_fetcher(
            "installations/" + date + "/", date=date)

    '''
        Paginated fetchers:
        Generators for paginated ListAPIViews by Serializer
    '''

    def __release_serializer_paginated_fetcher(self, url: str) -> Iterator[Release]:
        """
        Paginated fetcher for release serializer
        :param url: URL to fetch from
        """
        current_url = url
        while True:
            if current_url is None:
                break
            page_answer = self.__connection.get(request=current_url)
            page = page_answer.json()
            for release_desc in page['results']:
                if release_desc['prev_release'] is not None:
                    prev_release = self.__fetch_end_of_path_id(release_desc['prev_release'])
                else:
                    prev_release = None
                yield Release(
                    APIClient=self,
                    id=release_desc['id'],
                    url=release_desc['url'],
                    technology=self.technology(release_desc['technology']),
                    version=release_desc['version'],
                    prev_release=prev_release,
                    published=release_desc['published'],
                )
            current_url = page['next']

    def __sourcefile_serializer_paginated_fetcher(self, url: str) -> Iterator[SourceFile]:
        """
        Paginated fetcher for sourcefile serializer
        :param url: URL to fetch from
        """
        current_url = url
        while True:
            if current_url is None:
                break
            page_answer = self.__connection.get(request=current_url)
            page = page_answer.json()
            for sourcefile_desc in page['results']:
                yield SourceFile(
                    APIClient=self,
                    id=sourcefile_desc['id'],
                    url=sourcefile_desc['url'],
                    hash_sha1=sourcefile_desc['hash_sha1'],
                    hash_sha256=sourcefile_desc['hash_sha256'],
                    hash_md5=sourcefile_desc['hash_md5'],
                    hash_ssdeep=sourcefile_desc['hash_ssdeep']
                )
            current_url = page['next']

    def __vulnerability_serializer_paginated_fetcher(self, url: str) -> Iterator[Vulnerability]:
        """
        Paginated fetcher for vulnerability serializer
        :param url: URL to fetch from
        """
        current_url = url
        while True:
            if current_url is None:
                break
            page_answer = self.__connection.get(request=current_url)
            page = page_answer.json()
            for vuln_desc in page['results']:
                yield Vulnerability(
                    APIClient=self,
                    id=vuln_desc['id'],
                    url=vuln_desc['url'],
                    cve_number=vuln_desc['CVE_number'],
                    publish_date=vuln_desc['publish_date'],
                    last_modified=vuln_desc['last_modified'],
                    title=vuln_desc['title'],
                    type=vuln_desc['type'],
                    references=vuln_desc['references'],
                    fishy=vuln_desc['fishy']
                )
            current_url = page['next']

    def __installations_day_serializer_paginated_fetcher(self, url: str) -> Iterator[
        DateStatistics]:
        """
        Paginated fetcher for installations_day serializer
        :param url: URL to fetch from
        """
        current_url = url
        while True:
            if current_url is None:
                break
            page_answer = self.__connection.get(request=current_url)
            page = page_answer.json()
            for ds_desc in page['results']:
                yield DateStatistics(
                    APIClient=self,
                    date=ds_desc['date']
                )
            current_url = page['next']

    def __release_counter_release_serializer_paginated_fetcher(self, url: str, release: int) -> \
    Iterator[
        ReleaseCounter]:
        """
        Paginated fetcher for release_counter release serializer
        :param url: URL to fetch from
        :param release: Private key of associated release
        """
        current_url = url
        while True:
            if current_url is None:
                break
            page_answer = self.__connection.get(request=current_url)
            page = page_answer.json()
            for rc_desc in page['results']:
                if "observed_day" in rc_desc.keys():
                    date = rc_desc["observed_day"]
                else:
                    raise ValueError("ReleaseCounter needs release AND date information.")
                yield ReleaseCounter(
                    APIClient=self,
                    release=release,
                    observed_day=date,
                    count=int(rc_desc["count"])
                )
            current_url = page['next']

    def __release_counter_date_serializer_paginated_fetcher(self, url: str, date: str) -> Iterator[
        ReleaseCounter]:
        """
        Paginated fetcher for release_counter date serializer
        :param url: URL to fetch from
        :param date: Associated date in ISO-Format
        """
        current_url = url
        while True:
            if current_url is None:
                break
            page_answer = self.__connection.get(request=current_url)
            page = page_answer.json()
            for rc_desc in page['results']:
                release = None
                if "release" in rc_desc.keys():
                    release = self.__fetch_end_of_path_id(rc_desc["release"])
                if release is None:
                    raise ValueError("ReleaseCounter needs release AND date information.")
                yield ReleaseCounter(
                    APIClient=self,
                    release=release,
                    observed_day=date,
                    count=int(rc_desc["count"])
                )
            current_url = page['next']

    def __release_sourcefile_serializer_paginated_fetcher(self, url: str) -> Iterator[
        ReleaseSourcefile]:
        """
        Paginated fetcher for release_sourcefile serializer
        :param url: URL to fetch from
        """
        current_url = url
        while True:
            if current_url is None:
                break
            page_answer = self.__connection.get(request=current_url)
            page = page_answer.json()
            for rs_desc in page['results']:
                releaseSourcefile = None
                yield ReleaseSourcefile(
                    APIClient=self,
                    release=self.__fetch_end_of_path_id(rs_desc['release']),
                    sourcefile=self.__fetch_end_of_path_id(rs_desc['sourcefile']),
                    fully_qualified_name=rs_desc['fully_qualified_name']
                )
            current_url = page['next']

    def __fetch_technologies(self) -> Dict[str, Technology]:
        """
        Fetches the technologies/-Endpoint and returns found technologies as dict
        """
        technology_data = self.__connection.get(request="technologies/").json()
        technologies = {}
        for tech_desc in technology_data:
            technologies.update({tech_desc['tech_name']: Technology(
                APIClient=self,
                name=tech_desc['tech_name'],
                id=int(tech_desc['id']),
                releases_url=tech_desc['releases']
            )})
        return technologies

    def __fetch_end_of_path_id(self, url: str) -> Optional[int]:
        """
        Extracts an ID found as a last URL-path-part and returns it
        :param url: URL with ID in last path-part
        """
        urlparts = url.rstrip("/").split("/")
        try:
            return int(urlparts[len(urlparts) - 1])
        except ValueError:
            return None

    @staticmethod
    def destruct():
        """
        Destructs the ConnectionManager when the APIClient is destructed
        """
        ConnectionManager.destruct()



from __future__ import annotations

from datetime import datetime
from typing import Optional, Iterator

from weasel_client.resources.Resource import Resource


class Release(Resource):
    """
    Resource class for releases
    """
    _technology: Resource
    _version: str
    _prev_release: int = None
    _published: str

    def __init__(
            self,
            APIClient,
            id: int,
            url: str,
            technology: Resource,
            version: str,
            published: str,
            prev_release: int = None
    ):
        super().__init__(APIClient=APIClient, id=id, url=url)
        self._technology = technology
        self._version = version
        self._prev_release = prev_release
        self._published = published

    def technology(self) -> Resource:
        """
        Returns the release's technology
        """
        return self._technology

    def version(self) -> str:
        """
        Returns the release's version string
        """
        return self._version

    def prev_release(self) -> Optional[Release]:
        """
        Returns the release's previous release
        """
        if self._prev_release is not None:
            return self._client.release_detail(pk=self._prev_release)
        else:
            return None

    def published(self) -> datetime:
        """
        Returns the release's published timestamp
        :return:
        """
        return datetime.fromisoformat(self._published)

    def zipfile(self):
        """
        # Todo
        Should return the zipfile
        """
        pass

    def filelist(self) -> Iterator[Resource]:
        """
        Returns /releases/<release-ID>/filelist/ for the current release
        """
        yield from self._client.sourcefile_list(pk=self._id)

    def installations(self) -> Iterator[Resource]:
        """
        Returns /releases/<release-ID>/installations/ for the current release
        """
        yield from self._client.release_installations(pk=self._id)

    def vulnerabilities(self) -> Iterator[Resource]:
        """
        Returns /releases/<release-ID>/vulnerabilities/ for the current release
        """
        yield from self._client.release_vulnerabilities(pk=self._id)

    @staticmethod
    def from_id(APIClient, id: int) -> Optional[Release]:
        """
        Fetches a release-object with a given ID
        :param APIClient: client for API-requests
        :param id: id of the object to fetch
        """
        return APIClient.release_detail(pk=id)



from __future__ import annotations

from typing import Iterator, Optional

from weasel_client.resources.Resource import Resource


class SourceFile(Resource):
    """
    Resource class for SourceFiles
    """
    _file_data = None
    _hash_sha1 = None
    _hash_sha256 = None
    _hash_md5 = None
    _hash_ssdeep = None

    def __init__(
            self,
            APIClient,
            id,
            url,
            hash_sha1,
            hash_sha256,
            hash_md5,
            hash_ssdeep
    ):
        super().__init__(APIClient=APIClient, id=id, url=url)
        self._hash_sha1 = hash_sha1
        self._hash_sha256 = hash_sha256
        self._hash_md5 = hash_md5
        self._hash_ssdeep = hash_ssdeep

    def file(self):
        """
        # Todo
        Should return the file content
        """
        pass

    def hash_sha1(self):
        """
        Returns the sha1-hash in HEX format
        """
        return self._hash_sha1

    def hash_sha256(self):
        """
        Returns the sha256-hash in HEX format
        """
        return self._hash_sha256

    def hash_md5(self):
        """
        Returns the md5-hash in HEX format
        """
        return self._hash_md5

    def hash_ssdeep(self):
        """
        Returns the ssdeep-hash
        """
        return self._hash_ssdeep

    def releases(self) -> Iterator[Resource]:
        """
        Returns the associated releases
        """
        yield from self._client.sourcefile_releases(pk=self._id)

    @staticmethod
    def by_hash(APIClient, type: str, hash: str) -> Iterator[SourceFile]:
        """
        Fetches a SourceFile-object with a given hash
        :param APIClient: client for API-requests
        :param type: md5|sha1|sha256|ssdeep
        :param hash: hash in HEX format
        """
        return APIClient.sourcefile_byhash(type=type, hash=hash)

    @staticmethod
    def by_id(APIClient, id: int) -> Optional[SourceFile]:
        """
        Fetches a SourceFile-object with a given ID
        :param APIClient: client for API-requests
        :param id: id of the object to fetch
        """
        return APIClient.sourcefile_detail(pk=id)



from __future__ import annotations

from datetime import datetime, date
from typing import Optional, List, Iterator

from weasel_client.resources.Resource import Resource


class Vulnerability(Resource):
    """
    Resource class for Vulnerabilities
    """
    _cve_number: str
    _publish_date: str
    _last_modified: str
    _title: Optional[str]
    _type: Optional[str]
    _references: Optional[str]
    _fishy: bool

    def __init__(
            self,
            APIClient,
            id: int,
            url: str,
            cve_number: str,
            publish_date: str,
            last_modified: str,
            fishy: bool,
            title: Optional[str] = None,
            type: Optional[str] = None,
            references: Optional[str] = None,

    ):
        super().__init__(APIClient=APIClient, id=id, url=url)
        self._cve_number = cve_number
        self._publish_date = publish_date
        self._last_modified = last_modified
        self._title = title
        self._type = type
        self._references = references
        self._fishy = fishy

    def cve_number(self) -> str:
        """
        Returns the CVE Number
        """
        return self._cve_number

    def publish_date(self) -> date:
        """
        Returns the publish date
        """
        return date.fromisoformat(self._publish_date)

    def last_modified(self) -> datetime:
        """
        Returns the time the vulnerability was last modified
        """
        return datetime.fromisoformat(self._last_modified)

    def title(self) -> Optional[str]:
        """
        Returns the title
        """
        return self._title

    def type(self) -> Optional[str]:
        """
        Returns the vulnerability type
        """
        return self._type

    def references(self) -> Optional[List[str]]:
        """
        # Todo
        Should return the vulnerability references
        """
        if self._references is None:
            return None
        else:
            return None

    def fishy(self) -> bool:
        """
        Returns the fishy-flag
        """
        return self._fishy

    def found_in(self) -> Iterator[Resource]:
        """
        Returns the associated releases the vulnerability was found in
        """
        return self._client.vulnerability_found_in(pk=self._id)

    def fixed_by(self) -> Iterator[Resource]:
        """
        Returns the associated releases the vulnerability was fixed by
        """
        return self._client.vulnerability_fixed_by(pk=self._id)

    @staticmethod
    def from_id(APIClient, id: int) -> Optional[Vulnerability]:
        """
        Fetches a vulnerability-object with a given ID
        :param APIClient: client for API-requests
        :param id: id of the object to fetch
        """
        return APIClient.vulnerability_detail(pk=id)



from __future__ import annotations

from datetime import datetime
from typing import Optional, Iterator

from weasel_client.resources.Resource import Resource


class ReleaseSourcefile(Resource):
    """
    Resource class for ReleaseSourcefile
    """
    _release: int
    _sourcefile: int
    _fully_qualified_name: str

    def __init__(
            self,
            APIClient,
            release: int,
            sourcefile: int,
            fully_qualified_name: str
    ):
        super().__init__(APIClient=APIClient)
        self._release = release
        self._sourcefile = sourcefile
        self._fully_qualified_name = fully_qualified_name

    def release(self) -> Resource:
        """
        Returns the associated release
        """
        return self._client.release_detail(pk=self._release)

    def sourcefile(self) -> Resource:
        """
        Returns the associated sourcefile
        """
        return self._client.sourcefile_detail(pk=self._sourcefile)

    def fully_qualified_name(self) -> str:
        """
        Returns the fully_qualified_name of the sourcefile in the specific release
        """
        return self._fully_qualified_name



from weasel_client.resources.Resource import Resource


class Technology(Resource):
    _name = None
    _releases_url = None

    def __init__(self, APIClient, id, name, releases_url):
        super().__init__(APIClient=APIClient, id=id)
        self._name = name
        self._releases_url = releases_url

    def tech_name(self):
        """
        Returns the name of the technology
        """
        return self._name

    def releases(self):
        """
        Returns the associated releases
        """
        return self._client.release_tech_list(tech_name=self._name)

    def updates(self):
        """
        #Todo
        Should return the updates of the technology
        """
        pass

    @staticmethod
    def from_name(APIClient, name):
        """
        Fetches a Technology-object with a given name
        :param APIClient: client for API-requests
        :param name: name of the technology to fetch
        """
        return APIClient.technology(name=name)



from typing import Optional


class Resource:
    """
    Base class for resources fetched from the WEASEL-API
    """
    _client = None
    _id: Optional[int]
    _url: Optional[str]

    def __init__(self, APIClient, id: Optional[int] = None, url: Optional[str] = None):
        self._client = APIClient
        self._id = id
        self._url = url

    def id(self) -> Optional[int]:
        """
        Returns the ID
        """
        return self._id

    def pk(self) -> Optional[int]:
        """
        Alias for id
        """
        return self.id()

    def url(self) -> Optional[str]:
        """
        Returns the detail-endpoint URL of the resource
        """
        return self._url



import requests

from weasel_client.Exceptions import InvalidTokenException, ConnectionException


class ConnectionManager:
    """
    Singleton-Baseclass for the ConnectionManager
    """
    instance = None

    def __init__(self, token, url):
        if ConnectionManager.instance is None:
            ConnectionManager.instance = ConnectionManager.__ConnectionManager(token=token, url=url)

    @staticmethod
    def destruct():
        """
        Clears the instance on ConnectionManager-destruction
        """
        ConnectionManager.instance = None

    def __getattr__(self, name):
        return getattr(self.instance, name)

    class __ConnectionManager:
        """
        Singleton for connections to the WEASEL-API
        """
        token = None
        api_url = None

        def __init__(self, token, url):
            self.api_url = url
            self.token = token

        def get_token(self):
            """
            Returns the API-Token
            """
            return self.token

        def _set_token(self, token):
            """
            Sets the API-Token
            :param token: new API-Token
            """
            self.token = token

        def url(self):
            """
            returns the API-URL
            """
            return self.api_url

        def get(self, request):
            """
            Requests some result from the WEASEL-API and returns it
            :param request: Request-URL
            """
            if self.token is None:
                raise InvalidTokenException(
                    "No token given! Set the token using the 'set_token'-Method.")
            headers = {
                'Authorization': "Token " + self.token
            }
            if not request.startswith(self.api_url):
                request = self.api_url + request
            answer = requests.get(request, headers=headers)
            if answer.status_code == 200:
                return answer
            elif answer.status_code == 404 or answer.status_code == 500:
                raise ConnectionException("Could not connect to WEASEL API.")
            elif answer.status_code == 401:
                raise InvalidTokenException(
                    "Invalid token: '" + self.token + "'! Could not connect to WEASEL API.")
            else:
                raise ConnectionException(
                    "Unknown connection error. Status code: " + str(answer.status_code))

        def connected(self):
            """
            Simple check if the Connection Manager can connect to the WEASEL-API
            """
            try:
                self.get(request="")
                return True
            except (InvalidTokenException, ConnectionException):
                return False



from .DateStatistics import DateStatistics
from .Release import Release
from .ReleaseCounter import ReleaseCounter
from .ReleaseSourcefile import ReleaseSourcefile
from .SourceFile import SourceFile
from .Technology import Technology
from .Vulnerability import Vulnerability



import weasel_client.Exceptions
from weasel_client.APIClient import APIClient
from weasel_client.resources import DateStatistics
from weasel_client.resources import Release
from weasel_client.resources import ReleaseCounter
from weasel_client.resources import ReleaseSourcefile
from weasel_client.resources import SourceFile
from weasel_client.resources import Technology
from weasel_client.resources import Vulnerability



import copy
import datetime
import json
import lzma
import os
import requests
import shutil
import sqlite3


class ResultClient:
    def __init__(self, url, storage, pseudonym_path=None):
        self.url = url
        self.storage = storage
        self.index = set()
        self._get_index()
        self._pseudonym_path = pseudonym_path

        if pseudonym_path:
            if not os.path.exists(pseudonym_path):
                raise ValueError('Pseudonym path does not exist')
            self._sql_conn = sqlite3.connect(pseudonym_path)

    def _get_index(self):
        resp = requests.get(self.url + '/index.txt')
        resp.raise_for_status()
        self.index = set(d for d in resp.iter_lines(decode_unicode=True))

    def _get_cleartext_url(self, pseudonym):
        if self._pseudonym_path is None:
            raise ValueError('No path to pseudonym database given')

        cursor = self._sql_conn.cursor()
        res = cursor.execute('SELECT cleartext FROM pseudonyms_urls WHERE hash=?',
                             (pseudonym,)).fetchone()
        if res:
            return res[0]

    def _get_cleartext_domain(self, pseudonym):
        if self._pseudonym_path is None:
            raise ValueError('No path to pseudonym database given')

        cursor = self._sql_conn.cursor()
        res = cursor.execute('SELECT cleartext FROM pseudonyms_domains WHERE hash=?',
                             (pseudonym,)).fetchone()
        if res:
            return res[0]

    @property
    def results(self):
        return RangeSet(self)

    def get_day_result(self, day):
        return DayResults(self, day)


class RangeSet:
    _begin_year, _begin_month, _begin_day = None, None, None
    _end_year, _end_month, _end_day = None, None, None

    def __init__(self, client: ResultClient):
        self.client = client

    def range_from(self, year=None, month=None, day=None):
        c = copy.copy(self)
        if year:
            c._begin_year = year
        if month:
            c._begin_month = month
        if day:
            c._begin_day = day

        return c

    def range_to(self, year=None, month=None, day=None):
        c = copy.copy(self)
        if year:
            c._end_year = year
        if month:
            c._end_month = month
        if day:
            c._end_day = day

        return c

    def select_date(self, year, month, day):
        res = self.range_from(year, month, day)
        return res.range_to(year, month, day)

    def __iter__(self):
        first_day = "{:04d}-{:02d}-{:02d}".format(
            self._begin_year if self._begin_year else 0,
            self._begin_month if self._begin_month else 0,
            self._begin_day if self._begin_day else 0
        )

        last_day = "{:04d}-{:02d}-{:02d}".format(
            self._end_year if self._end_year else 9999,
            self._end_month if self._end_month else 12,
            self._end_day if self._end_day else 31
        )

        for day in sorted(self.client.index):
            if day < first_day:
                continue

            if day > last_day:
                return

            yield self.client.get_day_result(day)


class DayResults:
    def __init__(self, client: ResultClient, day):
        self.client = client
        self.day = day
        self.year, self.month, _ = self.day.split('-')

    def __str__(self):
        return "<DayResults({})>".format(self.day)

    @property
    def url(self):
        return f'{self.client.url}/{self.year}/{self.month}/{self.day}_Polecat.xz'

    def iter(self, path=None, stream=False):
        if stream:
            for res in self._stream_results():
                yield res
            return

        if path is None:
            path = self._prepare_path()

        if not os.path.isfile(path):
            path = self.download()

        with lzma.open(path) as fp:
            for line in fp:
                yield Result(client, json.loads(line))

    def _stream_results(self):
        resp = requests.get(self.url, stream=True)
        resp.raise_for_status()

        with lzma.LZMAFile(resp.raw) as fp:
            for line in fp:
                yield Result(self.client, json.loads(line))

    def download(self, path=None, override=False):
        if self.client.storage is None and path is None:
            raise ValueError('Path is set neither on function call nor in the ResultClient')

        resp = requests.get(self.url, stream=True)
        resp.raise_for_status()

        out_path = path if path else self._prepare_path()
        if os.path.exists(out_path) and not override:
            return out_path

        try:
            with open(out_path, 'wb') as fp:
                resp.raw.decode = True
                shutil.copyfileobj(resp.raw, fp)
        except:
            os.remove(out_path)
            raise

        return out_path

    def _prepare_path(self):
        try:
            os.mkdir(os.path.join(self.client.storage, self.year))
        except FileExistsError:
            pass

        try:
            os.mkdir(os.path.join(self.client.storage, self.year, self.month))
        except FileExistsError:
            pass

        return os.path.join(self.client.storage, self.year, self.month,
                            '{}_Polecat.xz'.format(self.day))


class Result:
    def __init__(self, client: ResultClient, raw: dict):
        self.client = client
        self.raw = raw

    @property
    def url(self):
        return self.client._get_cleartext_url(self.raw['url'])

    @property
    def domain(self):
        return self.client._get_cleartext_domain(self.raw['domain'])

    @property
    def matches(self):
        return self.raw['matches'] if 'matches' in self.raw else None

    @property
    def datetime(self):
        return datetime.datetime.strptime(self.raw['date'], '%Y-%m-%d %H:%M:%S.%f')


if __name__ == '__main__':
    # Create client connection and list available dates
    client = ResultClient('https://weasel.cs.uni-bonn.de/polecat/', 'cache',
                          'pseudonyms_prod.sqlite')
    print(client.index)

    # Iterate over a RangeSet
    for results in client.results.select_date(2020, 6, 1):
        print(results)
        i, j = 0, 0

        # Iter over day results
        for match in results.iter(stream=True):
            if match.url is None or match.domain is None:
                j += 1

            i += 1
            if i % 500000 == 0:
                print('Processed ', i)

        print('Results with missing pseudonyms: ', j)



import argparse
import datetime
import os
import sys

from weasel_client.raw_results.fetcher import ResultClient


def valid_date(s):
    try:
        return datetime.datetime.strptime(s, "%Y-%m-%d")
    except ValueError:
        msg = "Not a valid date: '{}'.".format(s)
        raise argparse.ArgumentTypeError(msg)


parser = argparse.ArgumentParser(
    description='Fetches RAW website scans and translates their pseudonyms back to cleartext domains and urls.')
parser.add_argument('--pseudonyms', type=str, help='Path to a pseudonym database file.')
parser.add_argument('--date-from', type=valid_date)
parser.add_argument('--date-to', type=valid_date)
parser.add_argument('--server', type=str, default='https://weasel.cs.uni-bonn.de/polecat/')
parser.add_argument('--storage', type=str, help='Path to a directory to store downloaded files.')
parser.add_argument('--download-only', action='store_true')
args = parser.parse_args()

if __name__ == '__main__':
    if args.download_only and args.storage is None:
        print('Storage path need to be defined if using download-only mode.')
        exit(-1)

    # Create client connection and list available dates
    client = ResultClient(args.server, args.storage, args.pseudonyms)

    qs = client.results
    if args.date_from:
        qs = qs.range_from(args.date_from.year, args.date_from.month, args.date_from.day)

    if args.date_to:
        qs = qs.range_to(args.date_to.year, args.date_to.month, args.date_to.day)

    # Iterate over a RangeSet
    for results in qs:
        if args.download_only:
            results.download()
            print('Downloaded {}-{}-{}'.format(results.year, results.month, results.day))
            continue

        # Iter over day results
        for match in results.iter(stream=args.storage is None):
            # Only return result of which we know the cleartext
            if match.url is not None and match.domain is not None:
                raw = match.raw
                raw['cleartext_url'] = match.url
                raw['cleartext_domain'] = match.domain
                print(raw)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['weasel_client',
 'weasel_client.raw_results',
 'weasel_client.resources',
 'weasel_client.scripts']

package_data = \
{'': ['*']}

install_requires = \
['python-dateutil>=2.8.1,<3.0.0', 'requests>=2.25.1,<3.0.0']

setup_kwargs = {
    'name': 'weasel-client',
    'version': '0.0.1',
    'description': '`weasel-client` is a python library to access the `weasel-api`.',
    'long_description': None,
    'author': 'Your Name',
    'author_email': 'you@example.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
